@@ -1,7 +1,6 @@
 /* Copyright (c) 2019 Joan Moreau <jom@grosjo.net>, see the included COPYING file */
 
 #include <xapian.h>
-#include <sqlite3.h>
 #include <cstdio>
 extern "C" {
 #include "fts-xapian-plugin.h"
@@ -15,11 +14,12 @@ extern "C" {
 #define XAPIAN_COMMIT_LIMIT 1000
 #define XAPIAN_COMMIT_TIMEOUT 300
 #define XAPIAN_WILDCARD "wldcrd"
-#define XAPIAN_EXPUNGE_SIZE 5
+#define XAPIAN_EXPUNGE_SIZE 2
+#define XAPIAN_EXPUNGE_HEADER 9
 
-#define HDRS_NB 10
-static const char * hdrs_emails[HDRS_NB] = { "uid", "subject", "from", "to",  "cc",  "bcc",  "messageid", "listid", "body", ""  };
-static const char * hdrs_xapian[HDRS_NB] = { "Q",   "S",       "A",    "XTO", "XCC", "XBCC", "XMID",      "XLIST",  "XBDY", "XBDY" }; 
+#define HDRS_NB 11
+static const char * hdrs_emails[HDRS_NB] = { "uid", "subject", "from", "to",  "cc",  "bcc",  "messageid", "listid", "body", "expungeheader",	""  };
+static const char * hdrs_xapian[HDRS_NB] = { "Q",   "S",       "A",    "XTO", "XCC", "XBCC", "XMID",      "XLIST",  "XBDY", "XEXP",		"XBDY" }; 
 
 static int verbose = 0;
 
@@ -38,9 +38,7 @@ struct xapian_fts_backend
 	char * old_guid;
 	char * old_boxname;
 
-	Xapian::Database * dbr;
 	Xapian::WritableDatabase * dbw;
-	sqlite3 * db_expunge;
 
 	long commit_updates;
 	long commit_time;
@@ -52,13 +50,6 @@ struct xapian_fts_backend
 
 };
 
-struct xapian_fts_expunge
-{
-	Xapian::docid did[XAPIAN_EXPUNGE_SIZE];
-	long uid[XAPIAN_EXPUNGE_SIZE];
-	int index;
-};
-
 struct xapian_fts_backend_update_context
 {
         struct fts_backend_update_context ctx;
@@ -86,15 +77,14 @@ static int fts_backend_xapian_init(struct fts_backend *_backend, const char **er
 	const char *const *tmp, *env;
 	long len;
 
-	backend->dbw = NULL;
-	backend->dbr = NULL;
 	backend->db = NULL;
+	backend->dbw = NULL;
+
 	backend->guid = NULL;
 	backend->path = NULL;
 	backend->old_guid = NULL;
 	backend->old_boxname = NULL;
 	backend->attachments = false;
-	backend->db_expunge = NULL;
 	backend->doexpunge = false;
 	verbose = 0;
 	backend->partial = 0;
@@ -176,14 +166,14 @@ static int fts_backend_xapian_init(struct fts_backend *_backend, const char **er
 
 static void fts_backend_xapian_deinit(struct fts_backend *_backend)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_deinit");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_deinit");
 
 	struct xapian_fts_backend *backend =
 		(struct xapian_fts_backend *)_backend;
 
 	if(verbose>1) i_info("Deinit %s",backend->path);
 
-	fts_backend_xapian_unset_box(backend);
+	if(backend->guid != NULL) fts_backend_xapian_unset_box(backend);
 
 	if(backend->old_guid != NULL) i_free(backend->old_guid);
 	backend->old_guid = NULL;
@@ -200,7 +190,7 @@ static void fts_backend_xapian_deinit(struct fts_backend *_backend)
 
 static int fts_backend_xapian_get_last_uid(struct fts_backend *_backend, struct mailbox *box, uint32_t *last_uid_r)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_get_last_uid");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_get_last_uid");
 
 	struct xapian_fts_backend *backend =
 		(struct xapian_fts_backend *)_backend;
@@ -213,7 +203,7 @@ static int fts_backend_xapian_get_last_uid(struct fts_backend *_backend, struct
 		return -1;
 	}
 
-	if(!fts_backend_xapian_check_read(backend))
+	if(!fts_backend_xapian_check_access(backend))
 	{
 		i_error("FTS Xapian: get_last_uid: can not open DB %s",backend->db);
 		return -1;
@@ -221,7 +211,7 @@ static int fts_backend_xapian_get_last_uid(struct fts_backend *_backend, struct
 
     	try
 	{
-		*last_uid_r = Xapian::sortable_unserialise(backend->dbr->get_value_upper_bound(1));
+		*last_uid_r = Xapian::sortable_unserialise(backend->dbw->get_value_upper_bound(1));
 	}
 	catch(Xapian::Error e)
 	{
@@ -238,7 +228,7 @@ static int fts_backend_xapian_get_last_uid(struct fts_backend *_backend, struct
 
 static struct fts_backend_update_context * fts_backend_xapian_update_init(struct fts_backend *_backend)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_update_context");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_update_context");
 
 	struct xapian_fts_backend_update_context *ctx;
 
@@ -249,7 +239,7 @@ static struct fts_backend_update_context * fts_backend_xapian_update_init(struct
 
 static int fts_backend_xapian_update_deinit(struct fts_backend_update_context *_ctx)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_update_deinit");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_deinit");
 
 	struct xapian_fts_backend_update_context *ctx =
 		(struct xapian_fts_backend_update_context *)_ctx;
@@ -265,7 +255,7 @@ static int fts_backend_xapian_update_deinit(struct fts_backend_update_context *_
 
 static void fts_backend_xapian_update_set_mailbox(struct fts_backend_update_context *_ctx, struct mailbox *box)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_update_set_mailbox");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_set_mailbox");
 
 	struct xapian_fts_backend_update_context *ctx =
 		(struct xapian_fts_backend_update_context *)_ctx;
@@ -277,33 +267,49 @@ static void fts_backend_xapian_update_set_mailbox(struct fts_backend_update_cont
 
 static void fts_backend_xapian_update_expunge(struct fts_backend_update_context *_ctx, uint32_t uid)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_update_expunge");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_expunge");
 
 	struct xapian_fts_backend_update_context *ctx =
 		(struct xapian_fts_backend_update_context *)_ctx;
 	struct xapian_fts_backend *backend =
 		(struct xapian_fts_backend *)ctx->ctx.backend;
 
-	if(!fts_backend_xapian_check_read(backend))
+	if(!fts_backend_xapian_check_access(backend))
 	{
-		i_error("FTS Xapian: Expunge UID=%d: Can not open db",uid);
+		i_error("FTS Xapian: Flagging UID=%d for expunge: Can not open db",uid);
 		return ;
 	}
 
     	try
 	{
-		if(verbose>0) i_info("FTS Xapian: Expunge UID=%d",uid);
+		if(verbose>0) i_info("FTS Xapian: Flagging UID=%d for expunge",uid);
 
 		XQuerySet * xq = new XQuerySet();
-		const char *u = t_strdup_printf("%d",uid);
+		char *u = i_strdup_printf("%d",uid);
 		xq->add("uid",u);
 
-		XResultSet * result=fts_backend_xapian_query(backend->dbr,xq,1);
-		
+		XResultSet * result=fts_backend_xapian_query(backend->dbw,xq,1);
+
+		i_free(u);
+
 		if(result->size>0)
 		{
 			Xapian::docid docid=result->data[0];
-			if(docid>0) fts_backend_xapian_add_expunge(backend,docid,uid);
+			if(docid>0) 
+			{
+				Xapian::Document doc = backend->dbw->get_document(docid);
+				try
+				{
+					doc.remove_term(hdrs_xapian[XAPIAN_EXPUNGE_HEADER]);
+				}
+				catch(Xapian::InvalidArgumentError e2)
+				{
+				}
+				u = i_strdup_printf("%s1",hdrs_xapian[XAPIAN_EXPUNGE_HEADER]);
+				doc.add_term(u);
+				backend->dbw->replace_document(docid,doc);	
+				i_free(u);
+			}
 		}
 
 		delete(result);
@@ -311,13 +317,13 @@ static void fts_backend_xapian_update_expunge(struct fts_backend_update_context
 	}
 	catch(Xapian::Error e)
 	{
-		i_error("FTS Xapian: (deleting) %s",e.get_msg().c_str());
+		i_error("FTS Xapian: Expunging UID=%d %s",uid,e.get_msg().c_str());
 	}
 }
 
 static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_context *_ctx, const struct fts_backend_build_key *key)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_update_set_build_key");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_set_build_key");
 
 	struct xapian_fts_backend_update_context *ctx =
 		(struct xapian_fts_backend_update_context *)_ctx;
@@ -411,7 +417,7 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 	}
 	if(i>=HDRS_NB) 
 	{ 
-		if(verbose>1) i_info("FTS Xapian: Unknown header '%s'",ctx->tbi_field); 
+		if(verbose>1) i_info("FTS Xapian: Unknown header (indexing) '%s'",ctx->tbi_field); 
 		i_free(ctx->tbi_field); 
 		ctx->tbi_field=NULL;
 		return FALSE; 
@@ -436,7 +442,7 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 
 static void fts_backend_xapian_update_unset_build_key(struct fts_backend_update_context *_ctx)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_update_unset_build_key");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_unset_build_key");
 
 	struct xapian_fts_backend_update_context *ctx =
 		(struct xapian_fts_backend_update_context *)_ctx;
@@ -451,7 +457,7 @@ static void fts_backend_xapian_update_unset_build_key(struct fts_backend_update_
 
 static int fts_backend_xapian_refresh(struct fts_backend * _backend)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_refresh");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_refresh");
 
 	struct xapian_fts_backend *backend =
 		(struct xapian_fts_backend *) _backend;
@@ -460,15 +466,14 @@ static int fts_backend_xapian_refresh(struct fts_backend * _backend)
         gettimeofday(&tp, NULL);
         long current_time = tp.tv_sec * 1000 + tp.tv_usec / 1000;
 
-	fts_backend_xapian_release(backend,"refresh");
-	fts_backend_xapian_commit(backend,"refresh", current_time);
+	fts_backend_xapian_release(backend,"refresh", current_time);
         
 	return 0;
 }
 
 static int fts_backend_xapian_update_build_more(struct fts_backend_update_context *_ctx, const unsigned char *data, size_t size)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_update_build_more");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_build_more");
 
 	struct xapian_fts_backend_update_context *ctx =
 		(struct xapian_fts_backend_update_context *)_ctx;
@@ -483,7 +488,7 @@ static int fts_backend_xapian_update_build_more(struct fts_backend_update_contex
 	icu::UnicodeString d2 = icu::UnicodeString::fromUTF8(sp_d);
 	if(d2.length() < backend->partial) return 0;
 
-	if(!fts_backend_xapian_check_write(backend))
+	if(!fts_backend_xapian_check_access(backend))
 	{
 		i_error("FTS Xapian: Buildmore: Can not open db");
 		return -1;
@@ -509,9 +514,7 @@ static int fts_backend_xapian_update_build_more(struct fts_backend_update_contex
 	if( (backend->commit_updates>XAPIAN_COMMIT_LIMIT) || ((current_time - backend->commit_time) > XAPIAN_COMMIT_TIMEOUT*1000) )
 	{
 		if(verbose>1) i_info("FTS Xapian: Refreshing after %ld ms and %ld updates...", current_time - backend->commit_time, backend->commit_updates);
-		fts_backend_xapian_release(backend,"refreshing");
-		//fts_backend_xapian_expunge(backend,"refreshing",XAPIAN_EXPUNGE_SIZE);
-		fts_backend_xapian_commit(backend,"refreshing", current_time);
+		fts_backend_xapian_release(backend,"refreshing", current_time);
 	}
     	
 	if(!ok) return -1;
@@ -520,26 +523,24 @@ static int fts_backend_xapian_update_build_more(struct fts_backend_update_contex
 
 static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_optimize");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_optimize");
 
 	struct xapian_fts_backend *backend =
 		(struct xapian_fts_backend *) _backend;
 
-	if(verbose>0) i_info("Optimize function");
-	
 	struct timeval tp;
         gettimeofday(&tp, NULL);
         long current_time = tp.tv_sec * 1000 + tp.tv_usec / 1000;
 	
-	fts_backend_xapian_expunge(backend,"optimize",0);
-        fts_backend_xapian_commit(backend,"optimize",current_time);
+	fts_backend_xapian_do_expunge(backend,"optimize",0);
+        fts_backend_xapian_release(backend,"optimize",current_time);
 
 	return 0;
 }
 
 static int fts_backend_xapian_rescan(struct fts_backend *_backend)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_rescan");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_rescan");
 
 	struct xapian_fts_backend *backend =
                 (struct xapian_fts_backend *) _backend;
@@ -556,7 +557,7 @@ static int fts_backend_xapian_rescan(struct fts_backend *_backend)
 
 static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbox *box, struct mail_search_arg *args, enum fts_lookup_flags flags, struct fts_result *result)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_lookup");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_lookup");
 
 	struct xapian_fts_backend *backend =
                 (struct xapian_fts_backend *) _backend;
@@ -569,12 +570,16 @@ static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbo
         gettimeofday(&tp, NULL);
         long current_time = tp.tv_sec * 1000 + tp.tv_usec / 1000;
 
-	fts_backend_xapian_commit(backend,"lookup", current_time);
+	Xapian::Database * dbr;
 
-	if(!fts_backend_xapian_check_read(backend))
+	i_array_init(&(result->maybe_uids),0);
+	i_array_init(&(result->scores),0);
+
+	if(!fts_backend_xapian_open_readonly(backend, &dbr))
         {
                 i_error("FTS Xapian: Lookup: Can not open db RO");
-                return -1;
+		i_array_init(&(result->definite_uids),0);
+                return 0;
         }
 
 	bool is_and=false;
@@ -592,20 +597,18 @@ static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbo
 	XQuerySet * qs = new XQuerySet(is_and,false,backend->partial);
 	fts_backend_xapian_build_qs(qs,args);
 
-	XResultSet * r=fts_backend_xapian_query(backend->dbr,qs);
+	XResultSet * r=fts_backend_xapian_query(dbr,qs);
 
 	long n=r->size;
 
 	i_array_init(&(result->definite_uids),r->size);
-        i_array_init(&(result->maybe_uids),0);
-	i_array_init(&(result->scores),0);
 
 	uint32_t uid;
 	for(long i=0;i<n;i++)
 	{
 		try
 		{
-			uid=Xapian::sortable_unserialise(backend->dbr->get_document(r->data[i]).get_value(1));
+			uid=Xapian::sortable_unserialise(dbr->get_document(r->data[i]).get_value(1));
 			seq_range_array_add(&result->definite_uids, uid);
 		}
 		catch(Xapian::Error e)
@@ -616,6 +619,9 @@ static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbo
 	delete(r);
 	delete(qs);
 
+	dbr->close();
+        delete(dbr);
+
 	/* Performance calc */
         if(verbose>0)
 	{
@@ -623,13 +629,12 @@ static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbo
         	long dt = tp.tv_sec * 1000 + tp.tv_usec / 1000 - current_time;
 		i_info("FTS Xapian: %ld results in %ld ms",n,dt);
 	}
-
 	return 0;
 }
 
 static int fts_backend_xapian_lookup_multi (struct fts_backend *_backend, struct mailbox *const boxes[], struct mail_search_arg *args, enum fts_lookup_flags flags, struct fts_multi_result *result)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_lookup_multi");
+	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_lookup_multi");
 
 	struct xapian_fts_backend *backend =
                 (struct xapian_fts_backend *) _backend;
